#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" flowtool-versioning cmdclass file. flowtool-versioning provides an
    automatic versioning system based on git tags. This is the command
    set, that it brings to add capabilities to your setup.py file to
    deal with git tags and handle the versioning. See the link for more
    information:

        https://github.com/isnok/py-flowlib
"""

# TODO: make these files pass pylint regularly
# pylint: disable=E0401,E1101,E0611

import os
import sys

from pprint import pformat
from os.path import join, exists, isfile, isdir, dirname, basename


### routines to initialize versioning


def find_in_parents(path, name):
    """ Return the nearest finding in the parent dirs of path,
       while searching for name. Returns None if no such parent dir exists.

        >>> find_in_parents('/tmp/', '__not_to_be_founfd_filename.txt__')
        >>> find_in_parents(dirname(__file__), basename(__file__)) == __file__
        True
    """

    while not isfile(join(path, name)):
        old = path
        path = dirname(path)
        if old == path:
            break
    else:
        return join(path, name)


def find_setup_cfg():
    """ Return the nearest directory in the parent dirs of path,
        that contains setup.cfg, or None if no such parent dir exists.

        >>> found = find_setup_cfg()
        >>> found is None or found.endswith('setup.cfg')
        True
    """
    for path in (os.getcwd(), dirname(__file__)):
        found = find_in_parents(path, 'setup.cfg')
        if found:
            return found

import sys
PYTHON = sys.version_info
configparser_module = 'ConfigParser' if PYTHON.major == 2 else 'configparser'
configparser = __import__(configparser_module)

def parse_setup_cfg():
    """ Find and parse the project setup.cfg that contains the versioning config.

        >>> hasattr(parse_setup_cfg(), 'has_option')
        True
    """
    parser = configparser.ConfigParser()
    parser.read(find_setup_cfg())
    return parser

def read_setup_cfg():
    """ Read relevant information from setup.cfg

        >>> all(x.endswith('.py') for x in read_setup_cfg())
        True
    """
    section = 'versioning'
    parser = parse_setup_cfg()

    source_versionfile = parser.get(section, 'source_versionfile')

    has_build = parser.has_option(section, 'build_versionfile')
    build_versionfile = parser.get(section, 'build_versionfile') if has_build else source_versionfile

    return source_versionfile, build_versionfile



def import_file(name, path):
    """ Import a python source file by its filesystem path.

        >>> from os.path import join, dirname
        >>> module = import_file('import_file_test', join(dirname(__file__), '__init__.py'))
    """

    module = None

    try: # py2
        import imp
        module = imp.load_source(name, path)
    except IOError:
        pass

    try: # py3.5
        import importlib.util
        spec = importlib.util.spec_from_file_location(name, path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
    except:
        pass

    try: # py3.3, py3.4
        from importlib.machinery import SourceFileLoader
        module = SourceFileLoader(name, path).load_module()
    except:
        pass

    return module


# this will be overriden later
def get_version(): return 'no_version'


def setup_versioning():
    """ Here some magic happens.

        >>> import sys
        >>> type(setup_versioning()) in (type(sys), type(None))
        True
    """

    global get_version

    source_versionfile, build_versionfile = read_setup_cfg()

    versionfile = import_file('_version', source_versionfile) if isfile(source_versionfile) else None

    get_version = versionfile.get_version if hasattr(versionfile, 'get_version') else get_version

    return versionfile


versionfile = setup_versioning()


def pretty_version_info(versionfile=versionfile):
    """ Testable body of a setuptools/distutils command.

        >>> pretty_version_info.VERSION_INFO = {}
        >>> pretty_version_info(pretty_version_info)
        '{}'
    """
    versioning_worked = versionfile is not None and hasattr(versionfile, 'VERSION_INFO')
    version_info = versionfile.VERSION_INFO if versioning_worked else None
    return pformat(version_info)


def print_version_info(self=None):
    """ Testable body of a setuptools/distutils command.

        >>> print_version_info()
        == Version-Config (setup.cfg):
        ...
    """
    parser = parse_setup_cfg()

    pretty_version_config = pformat(dict(parser.items('versioning')))
    print('== Version-Config (setup.cfg):\n' + pretty_version_config)
    print('== Version-Info:\n' + pretty_version_info())



### versioning config/setup done. start of setuptools helpers


from distutils.core import Command

def testable_do_nothing(self=None):
    """ A helper to increase test coverage.
        To be used on some classes as a method
        where a method override is required, but
        doesn't actually do anything.

        >>> testable_do_nothing()
    """

class cmd_version_info(Command):
    """ Version info command.

        Run `./setup.py version` to get detailed info on the latest version.
    """
    description = "show versioning configuration and current project version"
    user_options = []
    boolean_options = []

    # actually methods, but defined outside
    # the class for reuse- and testability.
    initialize_options = testable_do_nothing
    finalize_options = testable_do_nothing
    run = print_version_info



def bump_version(info):
    """ Bump a parsed version.

        >>> bump_version({'release':(8, 1)})
        {'release': (8, 2)}
        >>> bump_version({'release':(8, 1), 'post_release':0})['post_release']
        1
        >>> bump_version({'release':(8, 1), 'pre_release':('a', 0), 'dev_release':4})['dev_release']
        5
        >>> bump_version({'release':(8, 1), 'pre_release':('b', 0)})['pre_release']
        ('b', 1)
    """
    if 'dev_release' in info:
        info['dev_release'] += 1
    elif 'post_release' in info:
        info['post_release'] += 1
    elif 'pre_release' in info:
        stage, number = info['pre_release']
        info['pre_release'] = (stage, number + 1)
    else:
        segments = info['release']
        info['release'] = segments[:-1] + (segments[-1] + 1,)
    return info


def render_bumped(**kwd):
    """ Render the bumped version.
        >>> render_bumped(release=(1,2,3,4))
        '1.2.3.4'
        >>> render_bumped(release=(1,2,3,4), pre_release=('a', 5), post_release=6, dev_release=7, epoch=0)
        '0!1.2.3.4a5.post6.dev7'
        >>> render_bumped(release=(1,2,3), post_release=4, dev_release=5)
        '1.2.3.post4.dev5'
    """
    normalized = '.'.join(map(str, kwd['release']))
    if 'pre_release' in kwd:
        normalized += '%s%s' % kwd['pre_release']
    if 'post_release' in kwd:
        normalized += '.post' + str(kwd['post_release'])
    if 'dev_release' in kwd:
        normalized += '.dev' + str(kwd['dev_release'])
    if 'epoch' in kwd:
        normalized = '{}!{}'.format(kwd['epoch'], normalized)
    return normalized


def do_bump(self=None, test_data=None):
    """ Execute a version bump (if not testing)

        >>> do_bump(test_data={'dirt':'','tag_version':{'release':(1,1,2)},'prefix':''})
        == Next Version: {'release': (1, 1, 3)}
        == Tagging: 1.1.3
        >>> do_bump(test_data={'dirt':'XXX','tag_version':{'release':(1,1,2)},'prefix':''})
        Traceback (most recent call last):
        ...
        SystemExit: 1
    """
    vcs_info = versionfile.VERSION_INFO['vcs_info'] if test_data is None else test_data
    tag_info = bump_version(vcs_info['tag_version'])
    print('== Next Version: %s' % pformat(tag_info))
    if vcs_info['dirt']:
        print("==> Auto bump aborted due to dirty git repository.")
        sys.exit(1)
    tag = vcs_info['prefix'] + render_bumped(**tag_info)
    print('== Tagging: %s' % tag)
    test_data or os.system('git tag ' + tag)

class cmd_version_bump(Command):
    """ Version bump command.

        Run `./setup.py bump` to create a new git tag
        with the smallest version component increased
        by one.
    """
    description = "bump the (pep440) version by adding one to the smallest version component"
    user_options = []
    boolean_options = []

    initialize_options = testable_do_nothing
    finalize_options = testable_do_nothing
    run = do_bump


#class cmd_update_versionfile(Command):
    #description = "update the versioning"
    #user_options = []
    #boolean_options = []

    #def initialize_options(self):
        #pass

    #def finalize_options(self):
        #pass

    #def run(self):
        #print('== Updating file:\n%s' % source_versionfile)
        #from flowtool_versioning.dropins import version
        #versionfile = version.__file__
        #with open(versionfile, 'r') as f_in, open(source_versionfile, 'w') as f_out:
            #f_out.write(f_in.read())


# we override different commands for both environments
_sdist = _build_py = _upload = object

def import_commands_to_override(oldschool=False):
    """ Import the command classes to override either
        from setuptools or distutils.

        >>> import_commands_to_override()
        >>> import_commands_to_override(True)
    """
    global _sdist, _build_py, _upload
    from distutils.command.upload import upload as _upload

    if not oldschool and "setuptools" in sys.modules:
        from setuptools.command.sdist import sdist as _sdist
        from setuptools.command.build_py import build_py as _build_py
    else:
        from distutils.command.sdist import sdist as _sdist
        from distutils.command.build_py import build_py as _build_py

import_commands_to_override()

class cmd_build_py(_build_py):
    """ It seems as if build_py is executed when the distributed package is installed. """

    #def run(self):
        #_build_py.run(self)
        # now locate _version.py in the new build/ directory and replace
        # it with an updated value
        #deploy_to = build_versionfile
        #print("== Deploying %s" % deploy_to)
        #deploy_versionfile(deploy_to)


#if "cx_Freeze" in sys.modules:  # cx_freeze enabled?
    #from cx_Freeze.dist import build_exe as _build_exe

    #class cmd_build_exe(_build_exe):
        #def run(self):
            #root = get_root()
            #cfg = get_config_from_root(root)
            #versions = get_version()
            #target_versionfile = cfg.versionfile_source
            #print("UPDATING %s" % target_versionfile)
            #write_to_version_file(target_versionfile, versions)

            #_build_exe.run(self)
            #os.unlink(target_versionfile)
            #with open(cfg.versionfile_source, "w") as f:
                #LONG = LONG_VERSION_PY[cfg.VCS]
                #f.write(LONG %
                        #{"DOLLAR": "$",
                            #"STYLE": cfg.style,
                            #"TAG_PREFIX": cfg.tag_prefix,
                            #"PARENTDIR_PREFIX": cfg.parentdir_prefix,
                            #"VERSIONFILE_SOURCE": cfg.versionfile_source,
                            #})
    #cmds["build_exe"] = cmd_build_exe
    #del cmds["build_py"]


def add_to_sdist(self=None, base_dir=os.curdir, files=()):
    """ The custom part of the sdist command.

        >>> add_to_sdist(base_dir='/tmp')
        == Rendering:
        ...
        >>> add_to_sdist(base_dir='/tmp')
        == Rendering:
        ...
        >>> def boom(file=None):
        ...     raise OSError('File not found.')
        >>> import os
        >>> _exists = os.path.exists
        >>> os.path.exists = boom
        >>> add_to_sdist(base_dir='/tmp')
        == Rendering:
        ...
        >>> os.path.exists = _exists
    """
    # now locate _version.py in the new base_dir directory
    # (remembering that it may be a hardlink) and replace it with an
    # updated value

    self and _sdist.make_release_tree(self, base_dir, files)

    source_versionfile, build_versionfile = read_setup_cfg()
    target_versionfile = os.path.join(base_dir, build_versionfile)
    static_versionfile = versionfile.render_static_file() if versionfile else 'test_content'
    print("== Rendering:\n%s\n== To Versionfile: %s" % (static_versionfile, target_versionfile))

    try:
        # handles the hard link case correctly
        os.path.exists(target_versionfile) and os.unlink(target_versionfile)
        with open(target_versionfile, 'w') as fh: fh.write(static_versionfile)
    except:
        print("=== Could not render static _version.py to sdist!")

    self_target = join(base_dir, basename(__file__))
    print("== Updating: %s" % self_target)
    try:
        os.path.exists(self_target) or os.link(__file__, self_target)
    except OSError:
        print("=== Could not add %s to sdist!" % basename(__file__))

def sdist_run(self=None):
    """ A mere fake when run as a test... but 199% covered!

        >>> sdist_run()
    """
    if self: self.distribution.metadata.version = setup_versioning().get_version()
    if self: return _sdist.run(self)

class cmd_sdist(_sdist):

    run = sdist_run
    make_release_tree = add_to_sdist


def protected_upload(self=None):
    """ Allow only uploads with Python 3.
        I experienced problems earlier, when i uploaded packages built
        with python2. They containded .pyc files that made the distribution
        fail on python3 because of a bad magic number error. Since only I use
        this by now, i protect myself from this to happen again by this little
        mechanism. It may be removed or changed in the future.

        >>> protected_upload()
        ==> For backwards compatibility you should only upload packages built with Python 3 to PyPI.
    """
    if self and sys.version_info.major == 3: return _upload.run(self)
    print('==> For backwards compatibility you should only upload packages built with Python 3 to PyPI.')

class cmd_upload(_upload):
    description="Do the normal upload, but prevent pushing with Python2."
    run = protected_upload


def publish_code(self=None):
    """ push git and its tags

        >>> publish_code()
        === git push
        === pushing tags also

        TODO: do not use os.system, there are better ways...
    """
    print("=== git push")
    self and os.system('git push')
    print("=== pushing tags also")
    self and os.system('git push --tags')
    if self: return cmd_upload.run(self)


class cmd_release(cmd_upload):
    description="Do the protected upload, but push git things first"
    run = publish_code



def get_cmdclass():
    """ Return the custom setuptools/distutils subclasses.

        >>> sorted(get_cmdclass().keys())
        ['build_py', 'bump', 'release', 'sdist', 'upload', 'version']
    """
    cmds = dict(
        version=cmd_version_info,
        bump=cmd_version_bump,
        build_py=cmd_build_py,
        sdist=cmd_sdist,
        upload=cmd_upload,
        release=cmd_release,
    )
    return cmds


def main(noop=None):
    """ Prints the current version.

        >>> main(True)
        <BLANKLINE>
    """
    print(get_version() if not noop else '')

if __name__ == '__main__': main()



import os
import sys
import fnmatch

from flowtool.style import colors, echo, debug
from flowtool_git.common import local_repo, short_status

IGNORE_RECURSIVE = frozenset([
    '.git', '.tox', '.cache', 'build', 'dist', 'test', 'tests', 'venv',
])

def file_pattern_matcher(patterns=()):
    """ Create a matcher for multiple file patterns.

        >>> matches = file_pattern_matcher('*')
        >>> all(matches(x) for x in ('a', 'b.txt', 'c.py', 'd/b.wma'))
        True
    """
    if patterns == str(patterns):
        patterns = [patterns]

    def matches(filename):
        return patterns and any(fnmatch.fnmatch(filename, pat) for pat in patterns)

    return matches


def find_file_patterns_in_project(patterns=(), ignore_dirs=IGNORE_RECURSIVE, repo=None):
    """ Find file patterns in the repo recursively,
        ignoring some dirs only inside the repo.

        >>> len(find_file_patterns_in_project('*.py')) > 10
        True
    """

    if repo is None:
        repo = local_repo()

    matches = file_pattern_matcher(patterns)

    repo_root_segments = os.path.dirname(repo.git_dir).split(os.sep)

    def ignore_location(loc, dirs, files):
        inside = loc.split(os.sep)[len(repo_root_segments):]
        shall_ignore = ignore_dirs.intersection(inside)
        return bool(shall_ignore)

    result = []
    repo_root = os.path.dirname(repo.git_dir)
    for step in os.walk(repo_root):
        if not ignore_location(*step):
            loc, _, files = step
            relative_matches = [f for f in files if matches(f)]
            absolute_matches = [
                a for a in
                [os.path.join(loc, f) for f in files]
                if matches(a)
            ]
            result.extend(set(absolute_matches + [os.sep.join([loc, m]) for m in relative_matches]))
            debug.magenta(loc, matches)
    return result


def find_added_file_patterns(patterns=(), untracked_files=False, repo=None):
    """ Return the list of files that match the file name pattern
        and are added to the local git's index (but not as deleted).

        >>> bool(iter(find_added_file_patterns()))
        True
    """

    if repo is None:
        repo = local_repo()

    matches = file_pattern_matcher(patterns)

    untracked = 'yes' if untracked_files else 'no'
    added = []
    for line in short_status('--untracked-files=%s' % untracked):
        if line.on_index not in ' D':
            added.append(line.filename)

    return [f for f in added if matches(f)]


def find_changed_file_patterns(
        patterns=(),
        reference='origin/master',
        diff_filter='ACMRTUXB',  # exclude (only) (D)eleted by now
        repo=None,
    ):
    """ Return the list of files that match suffix and have changes
        in comparison to the reference_branch.

        diff_filter maps to the --diff-filter option of `git diff`:

        Select only files that are Added (A), Copied (C), Deleted (D),
        Modified (M), Renamed (R), have their type (i.e. regular file, symlink,
        submodule, ...) changed (T), are Unmerged (U), are Unknown (X), or have
        had their pairing Broken (B).

        >>> list(find_changed_file_patterns('*.not_here.orthere'))
        []
    """

    if repo is None:
        repo = local_repo()

    matches = file_pattern_matcher(patterns)

    try:
        changed = repo.git.diff(
            '--name-status',
            '--diff-filter=' + diff_filter,
            reference,
        ).split('\n')
    except:
        changed = ()
    parsed = [l.split('\t', 1)[1] for l in changed if l]

    return [f for f in parsed if matches(f)]



""" flowtool-githooks probing hooks.

    The functions in this module define the probing git hooks.
    The two probing hooks do either never or always fail, and
    dump the information they get (commandline args, stdin if
    available) to stdout. This makes them very practical when
    developing hooks, because they will show you the inputs
    of your hook functions.

    The two hooks here are kept to the minimum that is required
    to make a flowtool managed git hook. Implementing a setup
    method and defining the right entrypoints in setup.py (that
    is console_scripts for your hook, and flowtool_githooks.{name}
    with a setup function [for configs, etc]).



    >>> from click.testing import CliRunner
    >>> runner = CliRunner()

    >>> result = runner.invoke(probe, [])
    >>> result.exit_code
    0
    >>> result.output.startswith('probing-hook: ')
    True

    >>> result = runner.invoke(probe_fail, [])
    >>> result.exit_code
    -1
    >>> result.output.startswith('failing-probe-hook: ')
    True
"""
import os
import sys
import select
import click
from flowtool.style import echo, colors
from flowtool.python import read_stdin_nonblocking


def probe_setup(*args, **kwd):
    """ Shared setup function for the probing hooks.

        >>> probe_setup()
        probe-hook-setup: args=(), kwargs={}
        >>> probe_setup('install')
        probe-hook-setup: args=('install',), kwargs={}
        >>> probe_setup(cmd='uninstall')
        probe-hook-setup: args=(), kwargs={'cmd': 'uninstall'}
    """
    echo.white('probe-hook-setup: args=%s, kwargs=%s' % (args, kwd))


def dump_stdin_nonblocking():
    """ Dump stdin if there is any.

        >>> dump_stdin_nonblocking()
    """
    stdin = list(read_stdin_nonblocking(ignore_error=True))
    if stdin:
        stdin = ''.join(stdin)
        echo.cyan('>>>stdin>>>')
        echo.white(stdin, nl=False)
        echo.cyan('<<<stdin<<<')


@click.command()
@click.argument('args', nargs=-1)
def probe(args=()):
    """ A probing command function, to be run as a standalone (shell) command.
        It tries to take all input a git hook can get and display it.
    """
    try:
        echo.bold('probing-hook:', colors.cyan(sys.argv[0]), args)
        echo.white('running in:', colors.magenta(os.getcwd()))
        dump_stdin_nonblocking()
    except Exception as ex:
        echo.bold(type(ex), str(ex))


@click.command()
@click.argument('args', nargs=-1)
def probe_fail(args=(), noop=None):
    """ A probing hook that will always exit with a nonzero status.
        Usually that will be -1, because that's what click makes of
        an uncaught exception.
    """
    echo.red('failing-probe-hook:', colors.cyan(sys.argv[0]), args)
    dump_stdin_nonblocking()
    raise RuntimeError(42)



""" The githooks-runnner command takes care of most functions related
    to installing and manipulating the shipped runner script, that
    takes care of running multiple scripts inside one git hook.

    This file contains most of the functionality of the githooks-runner
    command. That is:

        - report the status (of installation)
        - installing and uninstalling the runner and scripts dir
        - toggling the executable bit of the runner (activating/deactivating the whole hook)
        - manually launching a hook script (i.e. for testing)

    >>> from click.testing import CliRunner
    >>> runner = CliRunner()

    >>> result = runner.invoke(runner_command, ())
    >>> result.exit_code
    0
    >>> result.output.startswith('git hooks status:')
    True
    >>> result = runner.invoke(runner_command, ('--status',))
    >>> result.exit_code
    0
    >>> result.output.startswith('git hooks status:')
    True
    >>> result = runner.invoke(runner_command, ('--status', 'xxx'))
    >>> result.exit_code
    0
    >>> result = runner.invoke(runner_command, ('--install',))
    >>> result.exit_code
    0
    >>> result = runner.invoke(runner_command, ('--activate',))
    >>> result.exit_code
    0
    >>> result = runner.invoke(runner_command, ('--deactivate',))
    >>> result.exit_code
    0
    >>> result = runner.invoke(runner_command, ('--remove',))
    >>> result.exit_code
    0
    >>> result = runner.invoke(runner_command, ('commit',))
    >>> result.exit_code
    1
    >>> result = runner.invoke(runner_command, ('--noop', 'commit-msg',))
    >>> result.exit_code
    0
"""
import os
import shutil
import filecmp

from os.path import exists, join, basename

import click

from flowtool.style import debug
from flowtool.style import colors, echo
from flowtool.python import containing
from flowtool.ui import abort
from flowtool.files import make_executable, make_not_executable, is_executable

from flowtool_git.common import local_repo

from flowtool_githooks.status import status as repo_status
from flowtool_githooks.manager import hook_specs, RUNNER


def run_githook(hook_name, noop=None):
    """ Execute a git hook.

        >>> run_githook('pre-cxmmit', noop=True)
        >>> install_runner('pre-commit')
        >>> run_githook('pre-commit', noop=True)
        Invoking pre-commit -> True ()
    """

    git_dir = local_repo().git_dir
    hook_file = os.sep.join([git_dir, 'hooks', hook_name])

    if not os.path.exists(hook_file):
        noop or abort('Hook does not exist: ' + hook_file)
        return
    elif not is_executable(hook_file):
        noop or abort('Hook script is not excutable:', hook_file)
        return

    spec = hook_specs[hook_name]
    echo.blue(
        colors.bold('Invoking'),
        colors.cyan(hook_name, bold=True),
        colors.cyan('->'),
        noop or colors.magenta(hook_file),
        spec.args
    )

    noop or os.execv(hook_file, spec.args or ('some_arg',))


@click.command()
@click.option('-i/-r', '--install/--remove', is_flag=True, default=None, help='Install or remove the runner script.')
@click.option('-a/-d', '--activate/--deactivate', is_flag=True, default=None, help='Manipulate executable bit of the runner script.')
@click.option('-s', '--status', is_flag=True, help='Print the status of the git hooks in the local repo.')
@click.option('-y', '--yes', is_flag=True, help='Automaticall answer yes to the backup question.')
@click.option('-n', '--noop', is_flag=True, help='Do not really take any action. Mainly for testing purposes.')
@click.argument('patterns', nargs=-1)
def runner_command(patterns=(), status=None, install=None, activate=None, yes=None, noop=None):
    """ Manage git hooks of the local repo. """

    hooks = containing(patterns, hook_specs)

    if status:
        return repo_status()

    if install is None and activate is None:

        if not hooks:
            return repo_status()

        elif len(hooks) > 1:
            noop or abort(
                'Too many hooks to run: ' +
                ', '.join(map(colors.cyan, hooks))
            )

        else:
            run_githook(hooks[0], noop=noop)

        return

    if not hooks:
        hooks = list(hook_specs)

    if install is not None:
        for hook in hooks:
            debug.white('install:', install, hook)
            if install is True:
                install_runner(hook, noop=noop, yes=yes)
            else:
                remove_runner(hook, noop=noop, yes=yes)

    if activate is not None:
        for hook in hooks:
            debug.white('activate:', activate, hook)
            if activate is True:
                activate_runner(hook, noop=noop)
            else:
                deactivate_runner(hook, noop=noop)


def remove_runner(hook_name, repo=None, noop=None, yes=None):
    """ Install the runner as a git hook.

        >>> remove_runner('xxx')
        Traceback (most recent call last):
        ...
        RuntimeError: not a supported git hook: 'xxx'
        >>> remove_runner('xxx', noop=True, yes=True)
    """

    if hook_name not in hook_specs and not noop:
        raise RuntimeError('not a supported git hook: %r' % hook_name)

    if repo is None:
        repo = local_repo()

    hook_file = join(repo.git_dir, 'hooks', hook_name)
    msg = ' '.join([
        colors.bold('Remove %s?' % hook_file),
    ])
    if yes or click.confirm(msg):
        noop or os.unlink(hook_file)


def install_runner(hook_name, repo=None, noop=None, yes=None):
    """ Install the runner as a git hook.

        >>> install_runner('xxx')
        Traceback (most recent call last):
        ...
        RuntimeError: not a supported git hook: 'xxx'
        >>> install_runner('xxx', noop=True, yes=True)
        >>> install_runner('pre-commit', noop=True)
        >>> install_runner('pre-commit', noop=True, yes=True)
    """

    if hook_name not in hook_specs and not noop:
        raise RuntimeError('not a supported git hook: %r' % hook_name)

    if repo is None:
        repo = local_repo()

    hook_file = join(repo.git_dir, 'hooks', hook_name)

    runner_file = RUNNER

    if exists(hook_file):
        if filecmp.cmp(runner_file, hook_file):
            make_executable(hook_file)
            return

        msg = ' '.join([
            'A script is already installed as the',
            colors.cyan(hook_name),
            'hook.\n',
            colors.bold('Do you want to remove it?'),
        ])
        if yes or (noop and click.confirm(msg)):
            noop or os.unlink(hook_file)

    noop or do_install(runner_file, hook_file)


def do_install(runner_file, hook_file, scripts_dir=None, quietly=None):
    """ Copy the runner_file into place and make it executable.
        Also create the scripts dir, if it does not already exist.

        >>> testfile = '/tmp/frobnication'
        >>> do_install(__file__, testfile, quietly=True)
        >>> do_install(__file__, testfile, quietly=True)
        >>> do_install(__file__, testfile, quietly=True)
        >>> os.path.isfile(testfile) and is_executable(testfile)
        True
        >>> os.path.isdir(testfile + '.d')
        True

        It will preserve the existing hook script, and move it to the
        scripts_dir, but it will not check if the runner is already in
        place as the git hook.
    """

    if scripts_dir is None:
        scripts_dir = hook_file + '.d'

    quietly or echo.white('Installing', runner_file, 'as', hook_file, '(scripts dir: %s).' % scripts_dir)

    exists(scripts_dir) or os.mkdir(scripts_dir)

    if exists(hook_file):

        backup_name = join(scripts_dir, basename(hook_file))
        while exists(backup_name):
            backup_name += '~'

        quietly or echo.white('Preserving original hook script as', backup_name)

        shutil.copyfile(hook_file, backup_name)
        make_not_executable(backup_name)
        os.unlink(hook_file)

    shutil.copyfile(RUNNER, hook_file)
    make_executable(hook_file)


def activate_runner(hook_name, repo=None, noop=None, yes=None):
    """ Activate a git hook (by making it executable).

        >>> activate_runner('xxx')
        Traceback (most recent call last):
        ...
        RuntimeError: not a supported git hook: 'xxx'
        >>> activate_runner('xxx', noop=True, yes=True)
        Activated xxx hook.
        >>> activate_runner('pre-commit', noop=True)
        Activated pre-commit hook.
        >>> activate_runner('pre-commit', noop=True, yes=True)
        Activated pre-commit hook.
    """

    if hook_name not in hook_specs and not noop:
        raise RuntimeError('not a supported git hook: %r' % hook_name)

    if repo is None:
        repo = local_repo()

    hook_file = join(repo.git_dir, 'hooks', hook_name)

    noop or make_executable(hook_file)
    echo.green('Activated', colors.cyan(hook_name), 'hook.')

def deactivate_runner(hook_name, repo=None, noop=None, yes=None):
    """ Deactivate a git hook (by making it not executable).

        >>> deactivate_runner('xxx')
        Traceback (most recent call last):
        ...
        RuntimeError: not a supported git hook: 'xxx'
        >>> deactivate_runner('xxx', noop=True, yes=True)
        Deactivated xxx hook.
        >>> deactivate_runner('pre-commit', noop=True)
        Deactivated pre-commit hook.
        >>> deactivate_runner('pre-commit', noop=True, yes=True)
        Deactivated pre-commit hook.
    """

    if hook_name not in hook_specs and not noop:
        raise RuntimeError('not a supported git hook: %r' % hook_name)

    if repo is None:
        repo = local_repo()

    hook_file = join(repo.git_dir, 'hooks', hook_name)

    noop or make_not_executable(hook_file)
    echo.green('Deactivated', colors.cyan(hook_name), 'hook.')



import os
import sys
import click
import filecmp

from collections import namedtuple
from pkg_resources import iter_entry_points

from flowtool.style import echo, colors
from flowtool.style import debug
from flowtool.files import is_executable, make_executable, make_not_executable

from flowtool_git.common import local_repo


# a reminder/description of what git hooks (can) take as input
#  - stdin not yet supported, not needed yet
#  - are there hooks whose stdout is used?
HookSignature = namedtuple('HookSignature', ['name', 'args', 'stdin'])

#  - this is currently just the hooks i want to focus on first
HOOK_SIGNATURES = [
    HookSignature('pre-commit', (), False),
    HookSignature('commit-msg', ('message_file',), False),
    HookSignature('pre-push', (), True),
]
hook_specs = {sig.name: sig for sig in HOOK_SIGNATURES}

RUNNER = os.sep.join([
    os.path.dirname(__file__), 'scripts', 'generic-hook-runner.sh'
])

InstalledHook = namedtuple('InstalledHook', ['name', 'active', 'file', 'is_runner', 'runner_dir'])

def script_group_name(hook_name):
    """ Derive the entry point group name from the git hook name.

        >>> script_group_name('pre-commit')
        'flowtool_githooks.pre_commit'
    """
    return 'flowtool_githooks.' + hook_name.replace('-', '_')

def get_script_entry_points(hook_name):
    """ Get the script entrypoints as a dict, using
        names as keys and entry points as values.

        >>> isinstance(get_script_entry_points('pre-commit'), dict)
        True
    """
    group = script_group_name(hook_name)
    result = {}
    itr = iter_entry_points(group)
    while True:
        try:
            e = next(itr)
            result[e.name] = e
        except StopIteration:
            break
        except:
            pass
    return result

def find_entry_scripts(hook_name):
    """ Find managed git hooks via the entry points for the respective hooks.
        This is done by intersecting the names in the virtual environments bin
        directory with the found entry points.

        Returns a dictionary with the absolute file names

        >>> find_entry_scripts('unknown-hook')
        {}
    """
    scripts = get_script_entry_points(hook_name)
    bindir = os.path.dirname(str(sys.executable))
    binscripts = set(scripts).intersection(os.listdir(bindir))
    entrypoint_scripts = {
        os.path.join(bindir, s): scripts[s]
        for s in binscripts
    }
    debug.bold('scripts for %r:' % hook_name, entrypoint_scripts)

    return entrypoint_scripts


def gather_file_hooks(repo=None):
    """ Gather information on installed git hook files.

        >>> type(gather_file_hooks()) == list
        True
    """

    if repo is None:
        repo = local_repo()

    hook_dir = os.path.join(repo.git_dir, 'hooks')
    files = os.listdir(hook_dir)
    hooks = [os.path.join(hook_dir,f) for f in files if f in hook_specs]
    found = []
    for filename in hooks:
        is_runner = filecmp.cmp(filename, RUNNER)
        runner_dir = filename + '.d'
        info = InstalledHook(
            name=os.path.basename(filename),
            active=is_executable(filename),
            file=filename,
            is_runner=is_runner,
            runner_dir=runner_dir if os.path.isdir(runner_dir) else None,
        )
        found.append(info)
    return sorted(found)


def gather_hooks(repo=None):
    """ Gather information on active git hooks.

        >>> gather_hooks() == gather_file_hooks()
        True
    """

    if repo is None:
        repo = local_repo()

    debug.cyan('Collecting information on installed hooks in', repo.git_dir)
    # config_hooks = gather_config_hooks(repo)
    file_hooks = gather_file_hooks(repo)
    return file_hooks



""" Shell Command Git Hooks.
"""
import os
import inspect

from functools import partial

from flowtool.style import echo, colors
from flowtool.execute import run_command
from flowtool.style import debug

from flowtool_git.common import local_repo, GitCommandError

from flowtool_githooks.managed_hooks.universal import make_check
from flowtool_githooks.managed_hooks.universal import ErroredCheck, CompletedCheck
from flowtool_githooks.managed_hooks.universal import UniversalGithook


def get_gitconfig_simple(repo=None, local=True):
    """ A very simple parser for the output of `git config --list`.

        >>> isinstance(get_gitconfig_simple(), dict)
        True
    """
    if repo is None:
        repo = local_repo()

    config_args = ('--list',)
    if local:
        config_args += ('--local',)

    dump = repo.git.config(*config_args)
    config = dict()
    for line in dump.split('\n'):
        key, value = line.split('=', 1)
        k1, k2 = key.split('.', 1)
        if k1 in config:
            config[str(k1)][str(k2)] = str(value)
        else:
            config[str(k1)] = {str(k2): str(value)}
    return config


class ConfiguredGithook(UniversalGithook):
    """ A git hook that brings some configuration stored in the
        local git repositories configuration. This means the file
        .git/config, accessible via `git config --local`.

        >>> tst = ConfiguredGithook()
        >>> tst.GITCONFIG_SECTION = 'GitConfigSection'
        Unsafe git config section 'GitConfigSection' corrected to 'gitconfigsection'.
        >>> tst.GITCONFIG_SECTION
        'gitconfigsection'
        >>> del tst.GITCONFIG_SECTION
    """

    @property
    def GITCONFIG_SECTION(self):
        if hasattr(self, '_GITCONFIG_SECTION'):
            return self._GITCONFIG_SECTION

    @GITCONFIG_SECTION.setter
    def GITCONFIG_SECTION(self, value):
        lower_value = value.lower()
        if lower_value != value:
            echo.yellow('Unsafe git config section %r corrected to %r.' % (value, lower_value))
        self._GITCONFIG_SECTION = lower_value

    @GITCONFIG_SECTION.deleter
    def GITCONFIG_SECTION(self):
        del self._GITCONFIG_SECTION

    GITCONFIG_DEFAULT = None

    def set_gitconfig(self, key, value):
        """ Set a key in the GITCONFIG_SECTION of this hook.

            >>> tst = ConfiguredGithook()

            Does not do anything if GITCONFIG_SECTION is not set:

            >>> tst.GITCONFIG_SECTION
            >>> tst.set_gitconfig('foo', 'bar')
            >>> tst.get_gitconfig()

            If the section is set (to a string hopefully), then we
            can store a value (also a string hopefully):

            >>> tst.GITCONFIG_SECTION = 'test-section-one'
            >>> tst.GITCONFIG_SECTION
            'test-section-one'
            >>> tst.set_gitconfig('foo', 'bar')
            >>> tst.get_gitconfig('foo')
            'bar'

            And we can also delete keys from there:

            >>> tst.del_gitconfig('foo')
            >>> tst.get_gitconfig()
            {}
            >>> tst.del_gitconfig()
        """
        if self.GITCONFIG_SECTION:
            gitconfig_key = '.'.join([self.GITCONFIG_SECTION, key])
            self.repo.git.config(gitconfig_key, value)


    def get_gitconfig(self, key=None):
        """ Get a key (or all) from the GITCONFIG_SECTION of this hook.

            >>> tst = ConfiguredGithook()
            >>> tst.GITCONFIG_SECTION = 'test-section-two'
            >>> tst.get_gitconfig()
            {}
            >>> tst.get_gitconfig('foo')
        """

        if self.GITCONFIG_SECTION:
            cfg_section = get_gitconfig_simple().get(self.GITCONFIG_SECTION, {})

            if key is None:
                return cfg_section

            return cfg_section.get(key)

    def del_gitconfig(self, key=None):
        """ Delete a key from, or the whole GITCONFIG_SECTION of this hook.

            >>> tst = ConfiguredGithook()
            >>> tst.GITCONFIG_SECTION = 'test-section-three'
            >>> tst.del_gitconfig()
            >>> tst.get_gitconfig()
            {}
            >>> tst.set_gitconfig('foo', 'bar')
            >>> tst.get_gitconfig()
            {'foo': 'bar'}
            >>> tst.set_gitconfig('bar', 'baz')
            >>> sorted(tst.get_gitconfig().keys())
            ['bar', 'foo']
            >>> tst.del_gitconfig('foo')
            >>> tst.get_gitconfig()
            {'bar': 'baz'}
            >>> tst.set_gitconfig('baz', 'foo')
            >>> tst.del_gitconfig()
            >>> tst.get_gitconfig()
            {}
        """
        if self.GITCONFIG_SECTION:

            def del_key(k):
                gitconfig_key = '.'.join([self.GITCONFIG_SECTION, k])
                self.repo.git.config('--unset-all', gitconfig_key)

            keys = [key]
            if key is None:
                keys = get_gitconfig_simple().get(self.GITCONFIG_SECTION, ())

            for key in keys:
                del_key(key)

            if key is None:
                try:
                    self.repo.git.config('--remove-section', self.GITCONFIG_SECTION)
                except GitCommandError:
                    pass


    def setup_gitconfig(self):
        """ Set up the git config section using the default values.

            >>> tst = ConfiguredGithook()
            >>> tst.GITCONFIG_SECTION = 'test-section-four'
            >>> tst.del_gitconfig()
            >>> tst.GITCONFIG_DEFAULT
            >>> data = {'foo': 'bar', 'bar': 'baz', 'baz': 'foo'}
            >>> tst.GITCONFIG_DEFAULT = data
            >>> tst.setup_gitconfig()
            >>> tst.get_gitconfig() == data
            True
            >>> tst.del_gitconfig()
        """
        if self.GITCONFIG_DEFAULT:
            for key, value in self.GITCONFIG_DEFAULT.items():
                self.set_gitconfig(key, value)


    def clean_gitconfig(self):
        """ Clean up the git config sections that are still set to the default values.

            >>> tst = ConfiguredGithook()
            >>> tst.GITCONFIG_SECTION = 'test-section-four'
            >>> tst.del_gitconfig()
            >>> tst.GITCONFIG_DEFAULT
            >>> data = {'foo': 'bar', 'bar': 'baz', 'baz': 'foo'}
            >>> tst.GITCONFIG_DEFAULT = data
            >>> tst.setup_gitconfig()
            >>> tst.set_gitconfig('foo', 'ooh')
            >>> tst.clean_gitconfig()
            >>> tst.get_gitconfig()
            {'foo': 'ooh'}
            >>> tst.del_gitconfig()
        """
        config = self.get_gitconfig()
        if self.GITCONFIG_DEFAULT and config:
            for key, value in config.items():
                if key in self.GITCONFIG_DEFAULT and self.GITCONFIG_DEFAULT[key] == value:
                    self.del_gitconfig(key)

    def hook_setup(self, cmd=None):
        """ Setup function for the hook

            >>> tst = ConfiguredGithook()
            >>> tst.GITCONFIG_DEFAULT = {'some': 'data'}
            >>> tst.hook_setup('install')
            >>> tst.get_gitconfig()
            >>> tst.hook_setup('uninstall')
            >>> tst.GITCONFIG_SECTION = 'test-section-five'
            >>> tst.hook_setup('install')
            >>> tst.get_gitconfig()
            {'some': 'data'}
            >>> tst.hook_setup('uninstall')
            >>> tst.get_gitconfig()
            {}
            >>> tst.del_gitconfig()
        """
        if cmd == 'install':
            self.setup_gitconfig()
        elif cmd == 'uninstall':
            self.clean_gitconfig()



class ConfigFileHook(ConfiguredGithook):
    """ A git hook that ships and utilizes a config file.

        The config file can be configured through a git config key.

        >>> tst = ConfigFileHook()
        >>> tst.GITCONFIG_SECTION = 'test-section-six'
        >>> tst.setup_configfile()
        >>> tst.setup_configfile()
        >>> tst.cleanup_configfile()
    """

    def __init__(self):
        super(ConfigFileHook, self).__init__()
        update =  {self.CONFIGFILE_GITCFGKEY: self._configfile_path()}
        if self.GITCONFIG_DEFAULT is None:
            self.GITCONFIG_DEFAULT = update
        else:
            self.GITCONFIG_DEFAULT.update(update)


    CONFIGFILE = '.universal.cfg'
    CONFIGFILE_GITCFGKEY = 'configfile'
    DEFAULT_CONFIGFILE = None

    def hook_setup(self, cmd=None):
        """ Setup function for the hook

            >>> tst = ConfigFileHook()
            >>> tst.GITCONFIG_DEFAULT = {'some': 'data'}
            >>> tst.hook_setup('install')
            >>> tst.get_gitconfig()
            >>> tst.hook_setup('uninstall')
            >>> tst.GITCONFIG_SECTION = 'test-section-wow'
            >>> tst.hook_setup('install')
            >>> tst.get_gitconfig()['some']
            'data'
            >>> tst.hook_setup('uninstall')
            >>> tst.del_gitconfig()
            >>> tst.get_gitconfig()
            {}
        """
        super(ConfigFileHook, self).hook_setup(cmd=cmd)
        if cmd == 'install':
            self.get_configfile(do_setup=True)
        elif cmd == 'uninstall':
            self.clean_gitconfig()

    def _configfile_path(self):
        configfile = os.path.join(
            os.path.dirname(self.repo.git_dir),
            self.CONFIGFILE,
        )
        return configfile

    def get_configfile(self, do_setup=None):
        """ Get the configuration value either from repo config
            or set it to its default in the repo config.

            >>> tst = ConfigFileHook()
            >>> tst.get_configfile().endswith(tst.CONFIGFILE)
            True
        """
        key = self.CONFIGFILE_GITCFGKEY
        value = self.get_gitconfig(key)

        if value is None:
            value = self._configfile_path()
            do_setup and self.set_gitconfig(key, value)

        if do_setup and not os.path.exists(value):
            self.create_configfile(value)

        return value

    def setup_configfile(self, config_file=None):
        """ Create a config file for the hook command,
            if it is not already there.

            >>> tst = ConfigFileHook()
            >>> tst.setup_configfile('/tmp/test')
            >>> tst.cleanup_configfile('/tmp/test')
        """
        if config_file is None:
            config_file = self.get_configfile(do_setup=True)
        if os.path.exists(config_file):
            debug.cyan(
                'universal-hook-setup:',
                os.path.basename(config_file),
                'exists',
            )
        else:
            self.create_configfile(config_file)

    def create_configfile(self, filename=None):
        """ Create the config file. For real.

            >>> from os.path import isfile
            >>> tst = ConfigFileHook()
            >>> tst.DEFAULT_CONFIGFILE = lambda: 'Hello World!'
            >>> tst.create_configfile('/tmp/_cfgfile_create_test')
            >>> isfile('/tmp/_cfgfile_create_test')
            True
            >>> tst.cleanup_configfile('/tmp/_cfgfile_create_test')
        """
        if filename is None:
            filename = self.get_configfile()
        default_config = self.DEFAULT_CONFIGFILE
        if default_config is not None:
            if str(default_config) == default_config:
                content = default_config
            else:
                content = default_config()
            with open(filename, 'w') as fh:
                fh.write(content)
            debug.cyan(
                'universal-hook-setup: created',
                os.path.basename(filename)
            )

    def cleanup_configfile(self, filename=None):
        """ Remove the config file. For real.

            >>> tst = ConfigFileHook()
            >>> tst.create_configfile()
            >>> tst.cleanup_configfile()
        """
        if filename is None:
            filename = self.get_configfile()
        if os.path.exists(filename):
            os.unlink(filename)


def capture_command(*cmdline, **kwd):
    """ Run command and return it's output.
        Issue a warning if the command is not installed.

        >>> capture_command('ls').returncode
        0
        >>> capture_command('ls').command
        ['ls']
        >>> try:
        ...     capture_command('_command_not_found_')
        ... except OSError:
        ...     print('Not found.')
        <BLANKLINE>
        ...
        Not found.
    """
    quiet_error = kwd['quiet_error'] if 'quiet_error' in kwd else False

    if len(cmdline) == 1:
        cmdline = cmdline[0]
    try:
        result = run_command(cmdline)
    except OSError as ex:
        msg = '\nAn Excepion occurred during command execution of: {}\nException: {}\n--> Is the command installed?'
        quiet_error or echo.yellow(msg.format(colors.cyan(' '.join(cmdline)), colors.red(repr(ex))))
        raise
    return result

make_command_check = partial(make_check, capture_command)


class ShellCommandHook(ConfigFileHook):
    """ A hook for abritrary shell commands.

        This hook introdces a consistent "pass" and "fail" logic for the checks.

        >>> tst = ShellCommandHook()
        >>> tst.CHECK_TOOL = 'true'
        >>> checks = [tst.make_check(x) for x in ('hello', 'world', '!!!')]
        >>> tst.adaptive_execution(checks=checks)
        == will check 3 files.
        == running: true hello 
        == true passed.
        == running: true world 
        == true passed.
        == running: true !!! 
        == true passed.
        0
        >>> tst.CHECK_TOOL = 'false'
        >>> checks = [tst.make_check(x) for x in ('hello', 'world', '!!!')]
        >>> tst.adaptive_execution(checks=checks)
        == will check 3 files.
        == running: false hello 
        == failed: false hello
        == Game Over: continues_used=1 returncode=1
        1
        >>> tst.CHECK_TOOL = 'false'
        >>> checks = iter([tst.make_check(x) for x in ('hello', 'world', '!!!')])
        >>> tst.adaptive_execution(checks=checks)
        == starting up: F
        == Game Over: continues_used=1 returncode=1
        == failed: false hello
        1
        >>> tst.CHECK_TOOL = 'true'
        >>> checks = iter([tst.make_check(x) for x in ('hello', 'world', '!!!')])
        >>> tst.adaptive_execution(checks=checks)
        == starting up: ...
        0
    """

    CHECK_TOOL = None
    RETURNCODE_ON_STDOUT = 0
    RETURNCODE_ON_STDERR = 0


    def _msg_hook_startup(self, checks=(), **kwd):
        msg = ('==', colors.cyan(self.NAME),) if hasattr(self, 'NAME') else ('==',)
        if hasattr(checks, '__len__'):
            if len(checks) > 1:
                msg += ('will check', colors.yellow(str(len(checks))), 'files.')
            elif len(checks) == 1:
                msg += ('will check', colors.yellow('1'), 'file.')
            else:
                return
        else:
            msg += ('starting up: ',)
        echo.bold(*msg, **kwd)


    def _msg_simple_check_start(self, check=None, **kwd):
        check_name = os.path.basename(check.args[0])
        args = ' '.join(check.args[1:]) if check.args else ''
        kwargs = check.kwargs if check.kwargs else ''
        msg = ('== running:', colors.cyan(check_name), args, kwargs)
        echo.white(*msg, **kwd)

    def _fmt_checked(self, outcome=None):
        check = outcome.check
        check_name = os.path.basename(check.args[0])
        if self.is_returncode(outcome):
            if type(outcome) is CompletedCheck:
                result = outcome.result
                command = (colors.yellow(os.path.basename(result.command[0])),) + result.command[1:]
                msg = ('== failed:', colors.yellow(' '.join(command)))
                if result.stdout or result.stderr:
                    msg += ('\n',)
                if result.stdout:
                    msg += (
                        # colors.cyan('\n> > > stdout > > >\n'),
                        '\n',
                        result.stdout,
                        # colors.cyan('\n< < < stdout < < <')
                    )
                if result.stderr:
                    msg += (
                        # colors.yellow('\n> > > stderr > > >\n'),
                        '\n',
                        colors.yellow(result.stderr),
                        # colors.yellow('\n< < < stderr < < <')
                    )
            elif type(outcome) is ErroredCheck:
                msg = ('== errored:', colors.yellow(check_name), outcome.exc_info[0], outcome.exc_info[1])
        else:
            msg = ('==', colors.cyan(check_name), 'passed.')
        return msg


    def _msg_generator_checked(self, outcome, **kwd):
        if type(outcome) is ErroredCheck:
            char = 'X'
        elif self.is_returncode(outcome):
            char = 'F'
        elif type(outcome) is CompletedCheck:
            char = '.'
        else:
            char = '?!'
        kwd.setdefault('nl', False)
        echo.bold(char, nl=False)


    def make_check(self, *args, **kwd):
        """ Make a check (combine function and args).

            >>> tst = ShellCommandHook()
            >>> len(tst.make_check().args)
            0
            >>> tst.CHECK_TOOL = 'flowtool'
            >>> len(tst.make_check().args)
            1
            >>> tst.CHECK_TOOL = ('flowtool', '--flow')
            >>> len(tst.make_check().args)
            2
        """
        if self.CHECK_TOOL is not None:
            if isinstance(self.CHECK_TOOL, (tuple, list)):
                tool_args = []
                for arg in self.CHECK_TOOL:
                    if arg == '>managed_config<':
                        tool_args.append(self.get_configfile())
                    else:
                        tool_args.append(arg)
                args = tuple(tool_args) + args
            else:
                args = (str(self.CHECK_TOOL),) + args

        return make_command_check(*args, **kwd)


    def is_returncode(self, result):
        """ Evaluate if the result adds to the returncode.

            >>> tst = ShellCommandHook()
            >>> tst.is_returncode(capture_command('ls'))
            0
        """
        if type(result) is ErroredCheck:
            return self.EXCEPTION_RETURNCODE
        elif type(result) is CompletedCheck:
            completed_command = result.result
        else:
            completed_command = result

        if completed_command.returncode:
            return completed_command.returncode
        elif completed_command.stderr and self.RETURNCODE_ON_STDERR:
            return self.RETURNCODE_ON_STDERR
        elif completed_command.stdout and self.RETURNCODE_ON_STDOUT:
            return self.RETURNCODE_ON_STDOUT
        else:
            return 0




"""
    This file was generated by flowtool-versioning.
    Homepage: https://github.com/isnok/py-flowlib
"""

VERSION_INFO = {'vcs_info': {'commit': '433e4a228dd5003fcfaadf1806f0a2be94ef4f59',
              'dirt': '',
              'latest_tag': 'flowtool-githooks-0.7.39',
              'latest_tag_commit': '433e4a228dd5003fcfaadf1806f0a2be94ef4f59',
              'latest_tag_version': '0.7.39',
              'prefix': 'flowtool-githooks-',
              'prefix_tag_distances': {'flowtool-githooks-0.7.36': 20,
                                       'flowtool-githooks-0.7.37': 18,
                                       'flowtool-githooks-0.7.38': 10,
                                       'flowtool-githooks-0.7.39': 0},
              'tag_version': {'normalized': '0.7.39',
                              'release': (0, 7, 39),
                              'version': '0.7.39'}},
 'version': '0.7.39'}

def get_version():
    global VERSION_INFO
    return VERSION_INFO.get('version', '$Format:%d %H$')



""" A universal python git hooks framework.

    Universal refers to the fact, that the given framework
    can be used for any sort of git hook (pre-commit, ...).

    This is not so hard to archieve, since the main differences
    between the hooks are merely the commandline arguments, and
    the fact, that some hooks get additional information via stdin.

    Also there is a difference in semantics (meaning/function of the
    hook), that is reflected, by allowing for different behaviour of
    the hook when called in different scenarios. Currently a universal
    hook has four modes of operation:

        - standalone:
            the hook is launched from the command line.
            (if possible) only now command-line arguments will be accepted
            there should be options to switch to the other run-modes
            i.e. a linter hook will in this mode check ALL available files
        - pre-commit:
            the hook is launched as a pre-commit hook
            i.e. a linter hook will in this mode check all files added to the commit
        - commit-msg:
            the hook is launched as a commit-msg hook
            i.e. a linter hook will in this mode check all files added to the commit
        - pre-push:
            the hook is launched as a pre-push hook
            i.e. a linter hook will in this mode check all files differing from origin/master


    >>> from click.testing import CliRunner
    >>> runner = CliRunner()

    >>> githook = UniversalGithook()
    >>> githook.generate_checks = lambda: [make_check(print_args, 'foo', bar='baz')]
    >>> result = runner.invoke(githook.click_command, [])
    >>> result.exception
    >>> result.exit_code
    0
    >>> output_lines = result.output.split('\\n')[:-1]
    >>> len(output_lines)
    4
    >>> output_lines[0] == '== will run 1 check.'
    True
"""
import os
import sys
import click

from collections import namedtuple

from flowtool.style import echo, colors
from flowtool.style import debug
from flowtool.python import read_stdin_nonblocking

from flowtool_git.common import local_repo, GitCommandError

from flowtool_githooks.discovering import find_file_patterns_in_project
from flowtool_githooks.discovering import find_changed_file_patterns
from flowtool_githooks.discovering import find_added_file_patterns


def print_args(*cmdline, **kwd):
    """ Print the args that are given to this function.
        Serves as a fallback for the check_func of a
        UniversalGithook.

        >>> print_args()
        dummy-check: () {}
    """
    echo.white('dummy-check:', cmdline, kwd)

def dummy_check(*args, **kwd):
    """ A dummy function, that will never fail.

        >>> dummy_check()
    """

Check = namedtuple('Check', ['func', 'args', 'kwargs'])
CompletedCheck = namedtuple('CompletedCheck', ['check', 'result'])
ErroredCheck = namedtuple('ErroredCheck', ['check', 'exc_info', 'returncode'])

def make_check(func=None, *args, **kwd):
    """ Create a Check object to be run by a UniversalGithook.

        >>> check = make_check(make_check, make_check)
        >>> check.func is check.args[0]
        True
    """
    return Check(func, args, kwd)


class UniversalGithook(object):
    """ The most simple form of a universal git hook.

        >>> tst = UniversalGithook()
        >>> tst.adaptive_execution(checks=iter([tst.make_check(x) for x in range(9)]))
        == starting up: .........
        0

    """

    FILE_PATTERNS = '*'

    EXCEPTION_RETURNCODE = -2
    CONTINUES = 0
    PROGRESSBAR_MIN_COUNT = 4
    SIMPLE_GENERATOR = False


    repo = local_repo()


    @classmethod
    def hook_setup(cls, cmd=None):
        """ Setup function for the hook

            >>> tst = UniversalGithook()
            >>> tst.hook_setup('install')
            >>> tst.hook_setup('uninstall')
        """

    def __init__(self):
        """ Create self.click_command. It shall be instance-bound,
            but click seemingly cannot decorate methods.
        """
        @click.command()
        @click.argument('args', nargs=-1)
        def click_command(args=()):
            sys.exit(self.adaptive_execution(args=args))
            # sys.exit(self.adaptive_execution(args=args))
        self.click_command = click_command


    RuntimeInfo = namedtuple(
        'RuntimeInfo',
        [
            'arg0',
            'args',
            'stdin',
            'run_mode',
        ],
    )

    arg0 = sys.argv[0]

    @property
    def args(self):
        if not hasattr(self, '_args'):
            self._args = tuple(sys.argv[1:])
        return self._args

    @args.setter
    def args(self, value):
        self._args = value

    @property
    def stdin(self):
        if not hasattr(self, '_stdin'):
            stdin = list(read_stdin_nonblocking(ignore_error=True))
            self._stdin = ''.join(stdin) if stdin else stdin
        return self._stdin

    @stdin.setter
    def stdin(self, value):
        self._stdin = value

    @property
    def run_mode(self):
        if not hasattr(self, '_run_mode'):
            arg0dir = self.arg0.split(os.sep)[-2]
            if arg0dir.endswith('.d'):
                self._run_mode = arg0dir[:-2]
            else:
                self._run_mode = 'standalone'
        return self._run_mode

    @run_mode.setter
    def run_mode(self, value):
        self._run_mode = value


    def collect_infos(self):
        """ Runtime information is collected through cached properties,
            to avoid unnecessary system calls. This function provides
            a combined test case and convenience function.

            >>> tst = UniversalGithook()
            >>> tst.stdin = 'foobar!baz'
            >>> infos = tst.collect_infos()
            >>> type(infos.args) is tuple
            True
            >>> infos.stdin
            'foobar!baz'
            >>> infos.run_mode
            'standalone'
            >>> tst.run_mode = 'something'
        """

        if not hasattr(self, 'args'):
            self.args = args

        info = self.RuntimeInfo(self.arg0, self.args, self.stdin, self.run_mode)
        return info

    def _check_func_name(self, func):
        return func.__name__ if hasattr(func, '__name__') else '<anon_check>'

    def _msg_simple_check_start(self, check=None, **kwd):
        check_name = self._check_func_name(check.func)
        args = '(%s)' % ', '.join(map(str,check.args)) if check.args else ''
        kwargs = check.kwargs if check.kwargs else ''
        msg = ('== running:', colors.cyan(check_name), args, kwargs)
        echo.white(*msg, **kwd)

    def _fmt_checked(self, outcome=None):
        check = outcome.check
        check_name = self._check_func_name(check.func)
        if self.is_returncode(outcome):
            msg = ('==', colors.cyan(check_name), 'errored.')
        else:
            msg = ('==', colors.cyan(check_name), 'passed.')
        return msg

    def _msg_simple_checked(self, outcome=None, **kwd):
        msg = self._fmt_checked(outcome)
        if self.is_returncode(outcome):
            echo.yellow(*msg, **kwd)
        else:
            echo.white(*msg, **kwd)

    def _msg_progressbar_failed(self, outcome=None, **kwd):
        msg = self._fmt_checked(outcome, **kwd)
        if self.is_returncode(outcome):
            echo.yellow('\n', *msg, **kwd)

    def _msg_generator_checked(self, outcome, **kwd):
        if type(outcome) is ErroredCheck:
            char = 'X'
        elif type(outcome) is CompletedCheck:
            char = '.'
        else:
            char = '?!'
        kwd.setdefault('nl', False)
        echo.bold(char, **kwd)

    def make_check(self, *args, **kwd):
        """ Make a check (combine function and args).

            >>> tst = UniversalGithook()
            >>> tst.check_func = 'test'
            >>> tst.make_check().func
            'test'
        """
        check_func = self.check_func if hasattr(self, 'check_func') else dummy_check
        return make_check(check_func, *args, **kwd)

    def generate_checks(self):
        """ Generate checks.

            >>> tst = UniversalGithook()
            >>> tst.run_mode
            'standalone'
            >>> len(tst.generate_checks()) > 42
            True
            >>> tst.run_mode = 'pre-commit'
            >>> lst = list(tst.generate_checks())
            >>> tst.run_mode = 'commit-msg'
            >>> lst == list(tst.generate_checks())
            True
            >>> tst.run_mode = 'pre-push'
            >>> bool(iter(tst.generate_checks()))
            True
        """
        if self.run_mode in ('pre-commit', 'commit-msg'):
            check_these = find_added_file_patterns(self.FILE_PATTERNS)
        elif self.run_mode in ('pre-push',):
            check_these = find_changed_file_patterns(self.FILE_PATTERNS)
        elif self.run_mode == 'standalone':
            check_these = find_file_patterns_in_project(self.FILE_PATTERNS)
        else:
            check_these = []

        return [self.make_check(f) for f in check_these]

    def run_check(self, check, **kwd):
        """ Run a check.

            >>> tst = UniversalGithook()
            >>> tst.run_check(make_check(print_args, 'Kowabunga!'))
            dummy-check: ('Kowabunga!',) {}
            ...
        """
        kwd.update(check.kwargs)
        try:
            check_result = check.func(*check.args, **kwd)
            return self.check_completed(check, check_result)
        except:
            return self.check_errored(check, sys.exc_info())

    def check_completed(self, check, result):
        return CompletedCheck(check, result)

    def check_errored(self, check, exc_info):
        return ErroredCheck(check, exc_info, self.EXCEPTION_RETURNCODE)

    def _msg_hook_startup(self, checks=(), **kwd):
        msg = ('==', colors.yellow(self.NAME),) if hasattr(self, 'NAME') else ('==',)
        if hasattr(checks, '__len__'):
            if len(checks) > 1:
                msg += ('will run', colors.green(str(len(checks))), 'checks.')
            elif len(checks) == 1:
                msg += ('will run', colors.green('1'), 'check.')
            else:
                return
        else:
            msg += ('starting up: ',)
        echo.bold(*msg, **kwd)

    def is_returncode(self, outcome):
        if type(outcome) is CompletedCheck:
            return 0
        else:
            return 1

    def execute_simple(self, checks=None, continues=None, **kwd):
        """ Simple procedure for hook execution.
            >>> tst = UniversalGithook()
            >>> tst.generate_checks = lambda: [make_check(print_args, 'Kowabunga!')]
            >>> tst.execute_simple()
            == will run 1 check.
            == running: print_args (Kowabunga!) 
            dummy-check: ('Kowabunga!',) {}
            == print_args passed.
            0
        """

        if checks is None:
            checks = self.generate_checks()

        if continues is None:
            continues = self.CONTINUES

        self._msg_hook_startup(checks)

        results = []
        fails = 0
        for check in checks:
            self._msg_simple_check_start(check)
            outcome = self.run_check(check, **kwd)
            self._msg_simple_checked(outcome)
            results.append(outcome)
            if self.is_returncode(outcome):
                fails += 1
                if fails >= continues:
                    return self.game_over(results, fails=fails)

        returncode = self.summarize(results, verbose=False)
        return returncode

    def execute_progressbar(self, checks=None, continues=None, **kwd):
        """ Procedure for hook execution using the click progressbar.

            >>> tst = UniversalGithook()
            >>> tst.generate_checks = lambda: [make_check(print_args, 'Kowabunga!')]
            >>> tst.execute_progressbar()
            == will run 1 check.
            <BLANKLINE>
            <BLANKLINE>
            dummy-check: ('Kowabunga!',) {}
            0
        """

        if checks is None:
            checks = self.generate_checks()

        if continues is None:
            continues = self.CONTINUES

        self._msg_hook_startup(checks)
        echo.white()

        results = []
        fails = 0
        with click.progressbar(checks) as bar:
            for check in bar:
                outcome = self.run_check(check, **kwd)
                results.append(outcome)
                if self.is_returncode(outcome):
                    self._msg_progressbar_failed(outcome)
                    fails += 1
                    if fails >= continues:
                        return self.game_over(results, fails=fails)

        returncode = self.summarize(results, verbose=False)
        return returncode

    def execute_generator(self, checks=(), continues=None, **kwd):
        """ Procedure for hook execution using the click progressbar.

            >>> tst = UniversalGithook()
            >>> tst.generate_checks = lambda: (i for i in [])
            >>> tst.execute_generator()
            <BLANKLINE>
            0
            >>> tst.generate_checks = lambda: [make_check(lambda x: '', 'Kowabunga!')]
            >>> tst.execute_generator()
            <BLANKLINE>
            0
        """

        if checks is None:
            checks = self.generate_checks()

        if continues is None:
            continues = self.CONTINUES

        self._msg_hook_startup(checks, nl=False)

        results = []
        fails = 0
        for check in checks:
            outcome = self.run_check(check, **kwd)
            self._msg_generator_checked(outcome)
            results.append(outcome)
            if self.is_returncode(outcome):
                fails += 1
                if fails >= continues:
                    echo.white('')
                    return self.game_over(results, fails=fails, verbose=True)
        echo.white('')
        returncode = self.summarize(results, verbose=True)
        return returncode

    def adaptive_execution(self, args=None, checks=None, **kwd):
        """ Auto select the execution style based on the capabilities of the checklist.

            >>> tst = UniversalGithook()
            >>> chk = make_check(lambda x: 'Kowa-', 'bunga!')
            >>> def checklist(n):
            ...     for _ in range(n):
            ...         yield chk
            ...
            >>> tst.generate_checks = lambda: []
            >>> tst.adaptive_execution()
            0
            >>> tst.generate_checks = lambda: [chk]
            >>> tst.adaptive_execution()
            == will run 1 check.
            == running: <lambda> (bunga!) 
            == <lambda> passed.
            0
        """
        if checks is None:
            checks = self.generate_checks()

        if hasattr(checks, '__len__'):
            if len(checks) < self.PROGRESSBAR_MIN_COUNT:
                return self.execute_simple(checks=checks, **kwd)
            else:
                return self.execute_progressbar(checks=checks, **kwd)
        else:
            if self.SIMPLE_GENERATOR:
                return self.execute_simple(checks=checks, **kwd)
            else:
                return self.execute_generator(checks=checks, **kwd)


    def game_over(self, results=(), fails=None, verbose=None):
        returncode = self.is_returncode(results[-1])
        echo.white('== Game Over:', 'continues_used=%s' % fails, 'returncode=%s' % returncode)
        if verbose:
            return self.summarize(results, verbose=verbose)
        return returncode

    def summarize(self, results=(), verbose=None):
        """ Summarize a list of CompletedCheck and ErroredCheck.

            >>> tst = UniversalGithook()
            >>> tst.summarize([CompletedCheck(make_check('func'), 0)])
            0
            >>> tst.summarize([ErroredCheck(make_check('func'), ('a', 'b', 'c'), 100)], verbose=True)
            == <anon_check> errored.
            1
        """
        returncode = 0
        fails = 0
        for outcome in results:
            ret = self.is_returncode(outcome)
            returncode |= ret
            if verbose and ret:
                self._msg_simple_checked(outcome)
        return returncode



""" Display the local git repositories hooks status.

    #>>> from click.testing import CliRunner
    #>>> runner = CliRunner()
"""
import os
import click

from flowtool.style import echo, colors
from flowtool.style import debug
from flowtool.files import is_executable
from flowtool_git.common import local_repo
from flowtool_githooks.manager import find_entry_scripts, gather_hooks

def status(repo=None, file_hooks=None):
    """ Draw a nice summary of the git hook status.

        >>> import os
        >>> if not os.path.exists('/tmp/test.d'): os.makedirs('/tmp/test.d')
        >>> from collections import namedtuple
        >>> Hook = namedtuple('InstalledHook', ['name', 'active', 'file', 'is_runner', 'runner_dir'])
        >>> status(file_hooks=[
        ...     Hook('hook_one', False, '/some/path', False, '/tmp/test.d'),
        ...     Hook('hook_two', True, '/some/other', True, '/tmp/test.d'),
        ... ])
        git hooks status:
        ...
    """

    if repo is None:
        repo = local_repo()

    if file_hooks is None:
        file_hooks = gather_hooks()

    echo.bold('git hooks status:')
    echo.white('git dir', repo.git_dir)
    for number, info in enumerate(file_hooks):

        if info.is_runner:
            effect = colors.bold
        else:
            effect = colors.white

        if info.active:
            color = colors.green
        else:
            color = colors.white

        hook_line = ' '.join([
            '\n==',
            colors.bold(colors.yellow('[{number}]')),
            '=',
            color(effect('{info.name}')),
            '=',
            color('enabled:{info.active:d}'),
            '=',
            effect('uptodate:{info.is_runner:d}'),
            '==',
            colors.magenta('{info.file}'),
        ])
        click.echo(hook_line.format(info=info, number=number+1))

        plugin_hooks = (os.path.basename(f) for f in find_entry_scripts(info.name))
        echo.white('Available:', colors.cyan(', '.join(sorted(plugin_hooks))))

        if info.runner_dir:
            scripts = sorted(os.listdir(info.runner_dir))
            if scripts:
                if info.active:
                    echo.white('Installed:')
                else:
                    echo.white('Installed, but disabled:')
            for script in scripts:
                fname = os.sep.join([info.runner_dir, script])
                if info.active and is_executable(fname):
                    color = echo.green
                else:
                    color = echo.white
                color('  - %s' % script, color=color)



""" Git hook configuration management.
    This file contains the main routines to configure git hooks interacitvely.

    >>> from click.testing import CliRunner
    >>> runner = CliRunner()

    >>> result = runner.invoke(manage_scripts, [])
    >>> result.exit_code
    2
    >>> result = runner.invoke(manage_scripts, ['--noop', '--remove', '--yes'])
    >>> result.exit_code
    2

    >>> result = runner.invoke(manage_scripts, ['pre-commit'])
    >>> result.exit_code in (0, -1)
    True
    >>> result = runner.invoke(manage_scripts, ['pre-commit', 'yaml'])
    >>> result.exit_code in (0, -1)
    True
"""
import os
from os.path import join, basename, dirname
import sys
import click

from flowtool.files import is_executable

from flowtool.style import echo, colors
from flowtool.style import debug
from flowtool.ui import abort
from flowtool.python import containing

from flowtool_git.common import local_repo

from flowtool_githooks.manager import get_script_entry_points
from flowtool_githooks.manager import hook_specs


def link_script(script_name, scripts_dir):
    """ Adds a script to a git hook runners dir.

        >>> link_script('test-script', '/tmp')
        >>> os.readlink('/tmp/test-script').startswith(dirname(sys.executable))
        True
        >>> os.unlink('/tmp/test-script')
    """
    script_src = join(dirname(sys.executable), script_name)
    script_dst = join(scripts_dir, script_name)
    os.symlink(script_src, script_dst)


def add_remove_scripts(hook, patterns, add, noop, scripts_dir, installed):
    """ Add or remove a script. This is called if --add or --remove are present.

        >>> add_remove_scripts('pre-commit', ('yaml'), True, True, '/tmp', ())
        >>> add_remove_scripts('pre-commit', ('yaml'), True, True, '/tmp', ('yaml',))
        >>> add_remove_scripts('pre-commit', ('yaml'), False, True, '/tmp', ())
        >>> add_remove_scripts('pre-commit', ('yaml'), False, True, '/tmp', ('yaml',))
    """
    if not patterns:
        abort('To add/remove you need to give some script name patterns.')

    available = get_script_entry_points(hook)
    matching = containing(patterns, available)

    if not matching:
        noop or abort('No scripts match %s for %s.' % (patterns, colors.cyan(hook)))

    debug.bold('add:', add, matching, installed, noop)

    if add is True:

        to_be_installed = set(matching).difference(installed)

        if not to_be_installed:
            msg = 'All of these are already in %s: %s'
            msg %= (colors.cyan(hook), ', '.join(map(colors.green, matching)))
            noop or abort(msg, returncode=0)

        for script in to_be_installed:
            noop or link_script(script, scripts_dir)

    elif add is False:

        to_be_removed = set(matching).intersection(installed)

        if not to_be_removed:
            msg = 'None of these are in %s: %s'
            msg %= (colors.cyan(hook), ', '.join(map(colors.green, matching)))
            noop or abort(msg, returncode=0)

        for script in to_be_removed:
            noop or os.unlink(join(scripts_dir, script))

    else:
        raise RuntimeError("Click betrayed us!")


@click.command()
@click.option('-a/-r', '--add/--remove', default=None, help='Wether the scripts should be added or removed.')
@click.option('-n', '--noop', is_flag=True, help='Do not really do anything. Mainly for testing purposes.')
@click.argument('hook', type=click.Choice(sorted(hook_specs)), nargs=1)
@click.argument('patterns', nargs=-1)
def manage_scripts(hook=None, patterns=(), add=None, noop=None, repo=None):
    """ Manage the scripts of a git hook runner. """

    if add is None and patterns:

        available = get_script_entry_points(hook)
        matching = containing(patterns, available)

        if matching:
            echo.white('Matching %s hook scripts:' % colors.cyan(hook))
            for script in matching:
                echo.green(' ->', script)
        else:
            echo.white('No matching scripts available.')
        return

    if repo is None:
        repo = local_repo()

    scripts_dir = join(repo.git_dir, 'hooks', hook + '.d')
    installed = os.listdir(scripts_dir)

    if add is not None:
        return add_remove_scripts(hook, patterns, add, noop, scripts_dir, installed)

    if installed:
        echo.white('Installed in %s:' % colors.cyan(scripts_dir))
        for idx, filename in enumerate(installed):
            cnt = idx + 1
            if is_executable(join(scripts_dir, filename)):
                echo.green(colors.white('%4d' % cnt), '+', filename)
            else:
                echo.white('%4d' % cnt, '+', filename)
    else:
        echo.white('Script dir %s is empty.' % colors.cyan(scripts_dir))

    installable = get_script_entry_points(hook)

    if installable:
        echo.white('Also available for %s:' % colors.blue(hook))
        for filename in installable:
            cnt += 1
            echo.cyan(colors.white('%4d' % cnt), '-', filename)
    else:
        echo.white('No further scripts available for %s.' % colors.cyan(hook))



import os
import sys
import click

from flowtool.files import is_executable, toggle_executable, find_parent_containing
from flowtool.style import echo, colors
from flowtool.ui import ask_choice

from flowtool_git.common import local_repo

from .manager import hook_specs

@click.command()
@click.option('-h', '--hook', type=click.Choice(hook_specs), default=None, help='Specify what hook to install.')
@click.option('-c', '--copy', is_flag=True, help='Copy the hook (default is symlinking).')
@click.argument('script', type=click.Path(exists=True, dir_okay=False, resolve_path=True))
def introduce_githook(script, hook=None, copy=None):
    """ Install a file (script) as git hook. """

    script_name = os.path.basename(script)
    if not hook:
        heading = ['Install', colors.cyan(script_name), 'into which git hook?']
        hook = ask_choice(
            heading=' '.join(heading),
            choices=hook_specs,
            question='Your choice',
        )

    echo.white(
        'Will install',
        colors.green(script_name),
        'as',
        colors.blue(hook),
        'using',
        colors.yellow('copy' if copy else 'symlink'),
        'installation method.',
    )

    hook_dest = os.path.join(
        local_repo().git_dir,
        'hooks',
        hook + '.d',
        script_name,
    )

    echo.white('Hook source:', colors.cyan(script))
    echo.white('Destination:', colors.cyan(hook_dest))

    if copy:
        os.link(script, hook_dest)
    else:
        os.symlink(script, hook_dest)

    echo.green('Done.')



#!/usr/bin/env python

#from distutils.core import setup
from setuptools import setup
from setuptools import find_packages

from versioning import get_cmdclass, get_version

import os


# General Info

setup_args = dict(
    name='flowtool-githooks',
    version=get_version(),
    cmdclass=get_cmdclass(),
    description='Automating tasks for git users.',
    author='Konstantin Martini',
    author_email='k@tuxcode.org',
    url='https://github.com/isnok/py-flowlib',
)


# Data Files

setup_args.update(
    packages=find_packages(),
    package_data={
        'flowtool_githooks': [
            'scripts/generic-hook-runner.sh',
        ],
    },
    zip_safe=False,
)


# Requirements

setup_args.update(
    install_requires=[
        'click',
        'flowtool-base',
        'flowtool-git',
        'gitpython',
        'pylint',
    ],
)


# Entry Points

setup_args.update(
    entry_points={
        'flowtool_commands': [
            'githooks-runner = flowtool_githooks.runner:runner_command',
            'githooks-introduce = flowtool_githooks.insert:introduce_githook',
            'githooks-manage = flowtool_githooks.config:manage_scripts',
        ],
        'console_scripts': [
            '_flowtool_githooks.probe = flowtool_githooks.probe_hooks:probe',
            '_flowtool_githooks.probe_fail = flowtool_githooks.probe_hooks:probe_fail',
        ],
        'flowtool_githooks.pre_commit': [
            '_flowtool_githooks.probe = flowtool_githooks.probe_hooks:probe_setup',
            '_flowtool_githooks.probe_fail = flowtool_githooks.probe_hooks:probe_setup',
        ],
        'flowtool_githooks.commit_msg': [
            '_flowtool_githooks.probe = flowtool_githooks.probe_hooks:probe_setup',
            '_flowtool_githooks.probe_fail = flowtool_githooks.probe_hooks:probe_setup',
        ],
        'flowtool_githooks.pre_push': [
            '_flowtool_githooks.probe = flowtool_githooks.probe_hooks:probe_setup',
            '_flowtool_githooks.probe_fail = flowtool_githooks.probe_hooks:probe_setup',
        ],
    },
)


# Techincal Details

setup_args.update(
    keywords=['git','flow','local','commandline'],
    platforms=['Debian/GNU Linux'],
    classifiers=[
        "Development Status :: 3 - Alpha",
        #"License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Topic :: System :: Software Distribution",
        "Topic :: System :: Systems Administration",
        "Topic :: Utilities",
        #"Topic :: Internet",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Environment :: Console",
    ],
)

if __name__ == '__main__': setup(**setup_args)
